@@ -11,10 +11,11 @@
 # See the "INSTALL.WEB" file for HTML fragments of files used by the
 # script (wait_t, done_t, and the main form itself), and crontab entries.
 # -----------------------------------------------------------------------
-my $VERSION = "1.53";
-# Last Modified: October 21, 2004
+my $VERSION = "3.00";
+# Last Major Modification: October 21, 2004
 # Changed aczone to aczoom
 # Updates by Avinash Chopde <avinash@acm.org> http://www.aczoom.com/
+# Updated: Aug 2007: added support for barcodegen using --tray-overlay
 # -----------------------------------------------------------------------
 # Created: March 2001, by Avinash Chopde <avinash@acm.org>  www.aczoom.com
 # -----------------------------------------------------------------------
@@ -67,8 +68,7 @@ my $VERSION = "1.53";
 #
 # Files/folders in $ROOTDIR:
 #   cdinsert.html [entry point, INSTALL.WEB has examples]
-#   wait_t.html
-#   done_t.html
+#   $INITIAL_TEMPLATE, $FORM_TEMPLATE, $WORKING_TEMPLATE, $DONE_TEMPLATE
 #   cdinsert.pl (copy to http:/cgi-bin)
 # Files/folders in any PATH dir:
 #   cdlabelgen
@@ -90,15 +90,19 @@ my $VERSION = "1.53";
 #-----------------------------------------------------------------------------
 use 5.005; # perl newer than 5.005 required
 use CGI 2.50 qw(escapeHTML); # 2.47 for upload, 2.50 for Vars
+use CGI::Carp qw(fatalsToBrowser);
 use Getopt::Std;
 use POSIX qw(floor);
 use File::Copy;
 use Socket qw(:DEFAULT :crlf);
+use HTML::Template;
+use HTML::FillInForm;
 #-----------------------------------------------------------------------------
-$start_time = time();
+$CGI::POST_MAX=1024 * 600;  # max size posts accepted, bytes
+# Note that the PDF file generated is around 4 times image sizes!
+
+$ENV{'PATH'}="/bin:/usr/bin:/usr/local/bin:/usr/local/netpbm/bin:/home/cgi/bin"; # security blanket (make sure all folders/files are non-writeable by others!)
 
-$SIG{HUP} = $SIG{INT} = $SIG{QUIT} = $SIG{PIPE} = $SIG{TERM} = \&sighandler;
-#-----------------------------------------------------------------------------
 # Following vars need to be set specifically for each site 
 $ROOTDIR = "/home/cgi/cdinsert"; # where all web files are kept
 
@@ -106,6 +110,86 @@ $ROOTDIR = "/home/cgi/cdinsert"; # where all web files are kept
 # the HTTP address is:
 # $ROOTHTTP = "/cgi/cdinsert"; # relative URL, absolute path
 
+$LOGFILE = "/home/cgi/tmp/weblog.txt";
+# log file - global script issues, not related to any one invocation
+# make sure this file exists, and is chmod 662 - writeable, but not readable
+# by others so casual web hackers can't read this directly using a browser.
+
+# for each invocation - semi-random name for files - gets some privacy, since
+# these folders and files are readable by the world.
+$WORKID = floor(rand(1e4)); 
+$WORKFILE = "cd" . $WORKID; # 'cd' followed by upto 4 digits
+$MSGFILENAME = "log$WORKID.txt";
+$SAVINPUT="inp$WORKID.txt"; # exact copy of form input string, saved for user
+$WEBDONE_HTML = "done$WORKID.html"; # complete HTML file (created in $WORKDIR)
+$SAVEFORM = "form" . $WORKID; # save form data here
+
+# .tmpl files are HTML::Template files - all in $ROOTDIR
+$INITIAL_TEMPLATE="initial.tmpl"; # first page, diplayed with form
+
+$FORM_TEMPLATE="form.tmpl"; # just the form
+
+$WORKING_TEMPLATE="working.tmpl"; # "working..." display template - this is
+# the HTML output of this script  to stdout, while it is doing the processing
+
+$DONE_TEMPLATE="done.tmpl"; # Template page with form (includes form.tmpl)
+
+# --- MRKR_* strings found in input HTML template files,
+# to be replaced with job-specific values. See sub replace_markers()
+# for list of all the marker tags - only non-HTML::Template tags is
+# SHOWSTATUS_HERE
+$SHOWSTATUS_HERE = "SHOWSTATUS_HERE\n"; # split WORKING_TEMPLATE on this line
+
+#--- tag printed in the HTML file, denotes start and end of user entered text
+# --------------
+$DEBUG = 0; # 0 no debug messages, 1 some messages, 2 more.
+# Debug messages may go to Apache error_log at the beginning, but
+# once the LOGFILE/MSGFILE is created, they will go there.
+
+delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};   # Make %ENV safer
+
+$sent_working_header = 0;
+
+#-----------------------------------------------------------------------------
+# initialize 
+
+$status_count = -2;
+$dontredirect = 0; # change to 1 to prevent auto redirect in WORKING_TEMPLATE
+$time_taken = 0;
+$time_taken_units = $time_taken;
+$errflag = 0;
+$errcmd = "";
+$hostname = "";
+$hostaddr = "";
+$form_html = ""; # filled in form, displayed on results page
+
+select((select(STDOUT), $| = 1)[0]); # autoflush
+
+$query = new CGI;
+
+# output HTML page headers, rest of HTML is output by copying WORKING_TEMPLATE
+print CGI::header();
+
+getopts('t:'); # -t <filename> is test mode. Filename is used as $incontents
+
+# --------------------------------
+# check if we need to do any work
+
+# if called without any form data (or -t test option), then just display
+# a page with the form and exit
+unless ($opt_t || $query->param()) {
+  my $tmpl = HTML::Template->new(filename => $INITIAL_TEMPLATE, path => [$ROOTDIR]);
+  print $tmpl->output();
+  exit(0);
+}
+
+#-----------------------------------------------------------------------------
+# start work, create work directory, etc
+
+$start_time = time();
+
+$SIG{HUP} = $SIG{INT} = $SIG{QUIT} = $SIG{PIPE} = $SIG{TERM} = \&sighandler;
+
 # for each invocation - semi-random name - gets some privacy, since
 # these folders are readable by the world.
 # $WORKDIRNAME = "cd" . floor(rand(100)) . "$$"; # keep it max 4 chars + $$
@@ -116,65 +200,15 @@ $WORKDIRNAME = &mktempdir("/home/cgi/tmp", $WORKDIRNAME, floor(rand(1e3))); # 1e
 $WORKDIR = "/home/cgi/tmp/$WORKDIRNAME"; 
 # "/usr/local/apache/htdocs/cgi" is symlink to "/home/cgi", so
 $WORKHTTP = "/cgi/tmp/$WORKDIRNAME"; 
-
-$LOGFILE = "/home/cgi/tmp/weblog.txt";
-# log file - global script issues, not related to any one invocation
-# make sure this file exists, and is chmod 662 - writeable, but not readable
-# by others so casual web hackers can't read this directly using a browser.
-
-# for each invocation - semi-random name for files - gets some privacy, since
-# these folders and files are readable by the world.
-$WORKID = floor(rand(1e4)); 
-$WORKFILE = "cd" . $WORKID; # 'cd' followed by upto 4 digits
-
-$MSGFILENAME = "log$WORKID.txt";
 $MSGFILE = "$WORKDIR/$MSGFILENAME";
 # STDOUT/STDERR messages collected here, different file for each invocation
 
-$WEBWAIT_THTML = "$ROOTDIR/wait_t.html"; # partial HTML file, template
-$WEBDONE_THTML = "$ROOTDIR/done_t.html"; # partial HTML file, template
-$WEBDONE_HTML = "done$WORKID.html"; # complete HTML file (created in $WORKDIR)
-
-$DEBUG = 0; # 0 no debug messages, 1 some messages, 2 more.
-# Debug messages may go to Apache error_log at the beginning, but
-# once the LOGFILE/MSGFILE is created, they will go there.
-
-$CGI::POST_MAX=1024 * 600;  # max size posts accepted, bytes
-# Note that the PDF file generated is around 4 times image sizes!
-
-$ENV{'PATH'}="/bin:/usr/bin:/usr/local/bin:/usr/local/netpbm/bin:/home/cgi/bin"; # security blanket (make sure all folders/files are non-writeable by others!)
-
 # following may be left alone, in most cases
 $ENV{'SHELL'} = "/bin/sh";
 $ENV{'TMPDIR'} = $WORKDIR;
 $ENV{'TEMP'} = $WORKDIR;
 $ENV{'TZ'} = "EST5EDT";
 
-delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};   # Make %ENV safer
-
-# constants ----------------------------------------------------------
-
-# --- MRKR_* variables are strings found in input HTML template files,
-# to be replaced with job-specific values
-$MRKR_HOSTNAME = "MRKR_HOSTNAME";
-$MRKR_WEBDONEHTML = "MRKR_WEBDONEHTML";
-$MRKR_PDF = "MRKR_PDF";
-$MRKR_POSTSCRIPT = "MRKR_POSTSCRIPT";
-$MRKR_GIF = "MRKR_GIF";
-$MRKR_INPUT = "MRKR_INPUT";
-$MRKR_SHOWSTATUS = "MRKR_SHOWSTATUS";
-$MRKR_SYSERROR = "MRKR_SYSERROR";
-$MRKR_SUCCESS = "MRKR_SUCCESS";
-$MRKR_ANYERROR = "MRKR_ANYERROR";
-$MRKR_GIFERROR = "MRKR_GIFERROR";
-$MRKR_MAKEGIF = "MRKR_MAKEGIF";
-$MRKR_CDLERROR = "MRKR_CDLERROR";
-$MRKR_MSGFILE = "MRKR_MSGFILE";
-$MRKR_TIMETAKEN = "MRKR_TIMETAKEN";
-$MRKR_DONTREDIRECT = "MRKR_DONTREDIRECT";
-
-$SAVINPUT="inp$WORKID.txt"; # exact copy of form input string, saved for user
-
 #--- executables - since am setting PATH, no need to give full path for each
 
 $REDIRECT="< /dev/null 2>&1";
@@ -202,22 +236,13 @@ $T_JPG2EPS_E = "jpeg2ps -r 72 -o $TIMAGEFILE.eps $TIMAGEFILE.jpg 2>&1";
 # jpegtran on every input JPG file to convert it to baseline JPEG
 $C_JPGTRAN_E = "jpegtran $CIMAGEFILE.jpg 2>&1 > $CIMAGEFILE.tmp && mv $CIMAGEFILE.tmp $CIMAGEFILE.jpg ";
 $T_JPGTRAN_E = "jpegtran $TIMAGEFILE.jpg 2>&1 > $TIMAGEFILE.tmp && mv $TIMAGEFILE.tmp $TIMAGEFILE.jpg ";
+# optional: if barcode generation has to be performed - requires
+# barcodegen - from GD::Barcode::Image to be available
+$BCODEFILE = "bcode$WORKID";
+$BCODE_E = "barcodegen --format=EPS --border=7 --write=$BCODEFILE.eps $REDIRECT";
+$CDL_BCODE_ARGS="--tray-overlay $BCODEFILE.eps --tray-overlay-scaleratio=1,-0.1,0.1";
 
 #-----------------------------------------------------------------------------
-# initialize 
-
-$status_count = -2;
-$dontredirect = ""; # make this 1 to prevent automatic redirect in WEBWAIT_THTML
-$time_taken = 0;
-$errflag = 0;
-$errcmd = "";
-$hostname = "";
-
-select((select(STDOUT), $| = 1)[0]); # autoflush
-
-# output HTML page headers, rest of HTML is output by copying WEBWAIT_THTML
-print CGI::header();
-
 open(LOGFILE, ">> $LOGFILE")  
   or myexit("Could not open log file ($LOGFILE): $!");
 select((select(LOGFILE), $| = 1)[0]); # autoflush
@@ -233,33 +258,57 @@ chdir $WORKDIR or myexit("cd ($WORKDIR) failed: $!");
 open(MSGFILE, ">> $MSGFILE")  
   or myexit("Could not open message file ($MSGFILE): $!");
 select((select(MSGFILE), $| = 1)[0]); # autoflush
-open(WEBWAIT_THTML, "< $WEBWAIT_THTML")  
-  or myexit("Could not open WEBWAIT_THTML file ($ROOTDIR/$WEBWAIT_THTML): $!");
-open(WEBDONE_THTML, "< $WEBDONE_THTML")
-  or myexit("Could not open WEBDONE_THTML file ($ROOTDIR/$WEBDONE_THTML): $!");
 open(WEBDONE_HTML, "> $WORKDIR/$WEBDONE_HTML")
   or myexit("Could not open WEBDONE_HTML file ($ROOTDIR/$WEBDONE_HTML): $!");
 
-$datestr = localtime($start_time);
-print MSGFILE "[$datestr]: Starting job in directory $WORKDIRNAME\n";
+open(SAVEFORM, ">> $SAVEFORM")  
+  or myexit("Could not open SAVEFORM file ($SAVEFORM): $!");
 
-# ----- file WEBWAIT opened, can send output...
+open(WORKING_TEMPLATE, "< $ROOTDIR/$WORKING_TEMPLATE")
+  or myexit("Could not open WORKING_TEMPLATE file ($WORKING_TEMPLATE): $!");
+
+$datestr = localtime($start_time);
+print MSGFILE "[$datestr] [pid $$]: Starting job in directory $WORKDIRNAME\n";
+
+# ----------------
+# send WORKING_TEMPLATE header before reading any input...
 # start_html etc is not used - instead, a template HTML file is read
 # in and output to STDOUT
-&print_cgihtml($MRKR_SHOWSTATUS); # keep copying until MRKR_SHOWSTATUS line...
-print &show_status();
+read(WORKING_TEMPLATE, $working_all_lines, $CGI::POST_MAX)
+	    or myexit("Could not read WORKING_TEMPLATE file ($WORKING_TEMPLATE): $!");
+($working_header, $working_trailer) = split(/$SHOWSTATUS_HERE/, $working_all_lines, 2);
+
+$working_header_tmpl = HTML::Template->new(scalarref => \$working_header,
+             path => [$ROOTDIR], die_on_bad_params => 0);
+
+&replace_markers($working_header_tmpl);
+
+# sent to client browser - half - the HTML page - after this, we will
+# print one line at a time, using show_status, for each command as it is
+# executed
+$working_header_tmpl->output(print_to => *STDOUT);
+$sent_working_header = 1;
+
+print &show_status("");
+
+# ----------------
+# allocate WORKING_TEMPLATE trailer after the header, as soon as possible,
+# in case of any calls to myexit(), so that trailer can be displayed
+if (!defined $working_trailer) {
+  myexit("Internal error - $WORKING_TEMPLATE missing SHOWSTATUS_HERE marker.");
+}
+$working_trailer_tmpl = HTML::Template->new(scalarref => \$working_trailer,
+             path => [$ROOTDIR], die_on_bad_params => 0);
 
 #----------------------------------------------------------------
 # collect user input
-$query = new CGI;
-
-getopt('t:'); # -t <filename> is test mode. Filename is used as $incontents
 
 if (! $opt_t ) {
     # in some cases, I've seen these variables come in as "undefined",
     # so to suppress perl -w warnings, using the || '' construct below
     $hostname = $query->remote_host() || '';
-    $useragent = $query->user_agent() || '';
+    $hostaddr = $query->remote_addr() || '';
+    $useragent = $query->user_agent() || 'CGI query - no useragent';
     $referer = $query->referer() || '';
     $intitle = $query->param('title') || '';
     $insubtitle = $query->param('subtitle') || '';
@@ -277,12 +326,17 @@ if (! $opt_t ) {
     $inmakegif = $query->param('makegif') || '';
     $ina4paper = $query->param('a4paper') || '';
     $insplititems = $query->param('splititems') || '';
-    $infile = $query->param('filename') || '';
+    $infilename = $query->param('filename') || '';
+    $infile= $query->upload('filename') || ''; # get file handle
+    $inbcodetype = $query->param('bcodetype') || '';
+    $inbcodetext = $query->param('bcodetext') || '';
 } else {
     $hostname = qq(testing mode "-t $opt_t");
-    $useragent = '';
+    $hostaddr = '';
+    $useragent = 'testing -t, no useragent';
     $referer = '';
-    $infile = '';
+    $infilename = '';
+    $infile = undef;
     $intitle = "Testing Title";
     $insubtitle = "Testing Subtitle";
     $incimage = 'mp3.eps';
@@ -297,6 +351,8 @@ if (! $opt_t ) {
     $ina4paper = 0;
     $inclogo = 1;
     $intlogo = 0;
+    $inbcodetype = '';
+    $inbcodetext = '';
     open(INPUT, "$opt_t")
       or myexit("Could not open input file (-t $opt_t) [Make sure full path given -t /tmp/x.txt etc]: $!");
     print MSGFILE "opened input file $opt_t\n" if ($DEBUG >= 1);
@@ -311,6 +367,8 @@ if (! $opt_t ) {
 $datestr = localtime($start_time);
 print LOGFILE "$WORKDIRNAME cdinsert [$datestr] $hostname - started\n";
 
+#----------------------------------------------------------------
+
 print MSGFILE "[$datestr] checking cgi_error...\n";
 $cgierror = $query->cgi_error(); # post too big, or user hit STOP, etc...
 # in some rare cases, cgi_error() takes over 2 minutes, not sure what
@@ -343,23 +401,43 @@ print MSGFILE "  referer = '$referer'\n";
 print MSGFILE "----------------------------------------------------\n";
 
 $gotstring = ($incontents =~ /\S+/);
-$gotfile = ($infile =~ /\S+/);
+$gotfile = ($infilename =~ /\S+/);
+$null_in_contents = -1;
 
 if ($gotfile) { # ignore $gotstring, file takes precedence
     if ($gotstring) {
 	print MSGFILE "** Warning: user entered text as well as filename, will append file to entered text.\n";
     }
+    # even when user types in non-existent file name, there is
+    # no error here, get an empty file. Should warn that
+    # file typed in upload field is empty
+    if (eof($infile)) {
+        myexit("Invalid upload file '$infilename' - no such file?");
+    }
     while (<$infile>) {
 	s/$CR?$LF/\n/; # correct end-of-line (variables from Socket package)
         $incontents .= $_; 
     }
-    print MSGFILE "... read in uploaded file: $infile\n";
+    print MSGFILE "... read in uploaded file: $infilename\n";
 
     # if the file is not text, but binary, reject it, stop script later..
-	# (stop later because file is saved before exiting, so we can debug)
-    $null_in_contents = ($incontents =~ /\000/);
+    # (stop later because file is saved before exiting, so we can debug)
+    # simple binary check - look for NULL character. Note that Unicode
+    # input text will have this null character in it - even if text is
+    # all just ASCII and user saves as Unicode....
+    $null_in_contents = index($incontents, "\000") + 1;
+
+    if (length($incontents) <= 0) {
+        myexit("Empty uploaded file '$infilename' - no data.");
+    }
 }
 
+# save users string before possible processing/changing, for debug.
+open(OUTPUT, "> raw$SAVINPUT")
+  or myexit("Could not open output file to save form input (raw$SAVINPUT): $!");
+print OUTPUT  $incontents;
+close OUTPUT;
+
 # some people post binary files here, and ghostscript gs hangs on
 # such text, so have to remove invalid characters
 # don't really know a sure-fire way of detecting binary files or
@@ -376,7 +454,8 @@ for (@items) {
     # remove all control chars, and all nulls
     s/[[:cntrl:]\000]//g;
     # restrict each line to max 256 characters
-    $_ = substr($_, 0, 256);
+    # no - don't mangle input, even if long lines
+    # $_ = substr($_, 0, 256);
 
     $incontents .= $_; 
     $incontents .= "\n";  # add back the end-of-line char
@@ -395,9 +474,12 @@ close OUTPUT;
 
 print MSGFILE "--> input text copied to $SAVINPUT.\n" if ($DEBUG >= 1);
 
+# save form data
+$query->save(\*SAVEFORM);
+
 # now exit if file was bad - NULLs in it, etc.
-if ($null_in_contents) {
-	myexit("\"$infile\" is not a ASCII text file. Only ASCII text content should be uploaded for list of items.");
+if ($null_in_contents > 0) {
+	myexit("'$infilename' - not ASCII or Latin1 text. Found null character in input ($null_in_contents).");
 }
 # ----------------------------------------------------------------------
 
@@ -410,6 +492,8 @@ $insplititems =~ /([\d\w\-\.]*)/; # no / allowed in name
 $insplititems = $1;
 $ina4paper =~ /([\d\w\-\.]*)/; # no / allowed in name
 $ina4paper = $1;
+$inbcodetype =~ /([\d\w\-\.]*)/; # no / allowed in name
+$inbcodetype = $1;
 
 # not really passing these as args, but perl -T complains, so clean them...
 $incdcase =~ /([\d\w\-\.]*)/; # no / allowed in name
@@ -433,11 +517,14 @@ $intitle =~ s/(\W)/\\$1/g;
 $insubtitle =~ /(.*)/; # yes, really need this.
 $insubtitle = $1;
 $insubtitle =~ s/(\W)/\\$1/g;
+$inbcodetext =~ /(.*)/; # yes, really need this.
+$inbcodetext = $1;
+$inbcodetext =~ s/(\W)/\\$1/g;
 
 print MSGFILE "after untaint: title($intitle) subtitle($insubtitle) clogo($inclogo) tlogo($intlogo)\n"
   if ($DEBUG >= 1);
 
-# read in any uploaded images
+# read in any uploaded images - cover image
 
 $datestr = localtime(time());
 print MSGFILE "[$datestr]: reading in any uploaded JPG files\n";
@@ -453,10 +540,10 @@ if ($incimagefile) {
     close($incimagefile);
 
     &do_cmd($C_JPGTRAN_E) || &do_cmd($C_JPG2EPS_E);
-    $dontredirect = "1" if ($errflag); # important command failed
+    $dontredirect = "1" if ($errflag); # above command failed
     $incimage = $errflag ? "": "$CIMAGEFILE.eps";
 }
-# read in any uploaded images
+# read in any uploaded images - tray image
 if ($intimagefile) {
     if ($intimage) {
 	print MSGFILE "** Warning: user selected built-in Tray Image and uploaded Image, ignoring built-in.\n";
@@ -468,7 +555,7 @@ if ($intimagefile) {
     close($intimagefile);
 
     &do_cmd($T_JPGTRAN_E) || &do_cmd($T_JPG2EPS_E);
-    $dontredirect = "1" if ($errflag); # important command failed
+    $dontredirect = "1" if ($errflag); # above command failed
     $intimage = $errflag ? "" : "$TIMAGEFILE.eps" ;
 }
 
@@ -490,6 +577,22 @@ if ($ina4paper) {
   $ina4paper = "-y 0.8" if ($incdcase =~ /^(dvd)|(envelope)/);
 }
 #-----------------------------------------------------------------------------
+# check if barcode generation has to be performed - this requires
+# barcodegen - from GD::Barcode::Image to be available
+if ($inbcodetype && $inbcodetext && !$errflag) {
+    $datestr = localtime(time());
+    print MSGFILE "[$datestr]: create barcode\n";
+    $BCODE_E .= " --type '$inbcodetype' $inbcodetext";
+    &do_cmd($BCODE_E);
+    if ($errflag) {
+	$dontredirect = "1"; # barcodegen failed, don't redirect
+
+	# this command is not important so reset error state,
+	# to allow other commands to proceed
+	$errflag--;
+    }
+}
+#-----------------------------------------------------------------------------
 
 $datestr = localtime(time());
 print MSGFILE "[$datestr] $hostname :: Starting programs...\n";
@@ -520,6 +623,7 @@ push(@cmdargs, "--create-dvd-outside") if ($incdcase =~ /^dvdoutside/);
 push(@cmdargs, "-p") if (! $inscaleitems); 
 push(@cmdargs, "-b") if ($innotrayhd);
 push(@cmdargs, $ina4paper) if ($ina4paper);
+push(@cmdargs, $CDL_BCODE_ARGS) if (-s "$BCODEFILE.eps");
 
 # if number of items is very large, print some items on the cover also
 push(@cmdargs, "-v " . int($num_items/2)) if ($num_items > 250 || $insplititems);
@@ -569,6 +673,7 @@ $end_time = time();
 #NOTUSED $processing_time_taken = $end_time - $start_processing_time;
 $receive_time_taken = $start_processing_time - $start_time;
 $time_taken = $end_time - $start_time;
+$time_taken_units = ($time_taken <= 1) ? "1 second" : "$time_taken seconds";
 $datestr = localtime($end_time);
 $datestr = "$WORKDIRNAME cdinsert [$datestr] took $time_taken secs ";
 $datestr .= "(download $receive_time_taken) " if ($gotfile || $incimagefile || $intimagefile);
@@ -580,34 +685,23 @@ print MSGFILE "----------------------------------------------------\n";
 
 # copy output HTML page show output or error, as appropriate
 
-while (<WEBDONE_THTML>) {
+# add the form to the done file, with filled in values
+$form_tmpl = HTML::Template->new(filename => $FORM_TEMPLATE, path => [$ROOTDIR]);
+$form_html = $form_tmpl->output();
+$form_fillin = HTML::FillInForm->new();
+$form_html = $form_fillin->fill(scalarref => \$form_html, fobject => $query);
 
-    if ($errflag) {
-      # depending on error, uncomment particular lines in WEBDONE_THTML
-      next if (&check_marker($_, $MRKR_ANYERROR));
-      if ($errcmd =~ /$CDLBL_E/) {
-	next if (&check_marker($_, $MRKR_CDLERROR));
-      } elsif ($errcmd =~ /$PGM2GIF_E/) {
-	next if (&check_marker($_, $MRKR_GIFERROR));
-      }
-      next if (&check_marker($_, $MRKR_SYSERROR));
-      # #errcmd not CDLBL or PGMGIF
-    } else { 
-      # no error, uncomment the SUCCESS lines
-      next if (&check_marker($_, $MRKR_SUCCESS));
-    }
-
-    next if ($inmakegif && &check_marker($_, $MRKR_MAKEGIF) && !$errflag);
-
-    print WEBDONE_HTML &replace_markers($_);
-}
+# complete rest of HTML from this CGI script, will redirect to WEBDONE
+# copy output HTML page - show output or error, as appropriate
+$done_tmpl = HTML::Template->new(filename => $DONE_TEMPLATE,
+             path => [$ROOTDIR], die_on_bad_params => 0);
+&replace_markers($done_tmpl);
+$done_tmpl->output(print_to => *WEBDONE_HTML);
 close(WEBDONE_HTML);
 
 # complete rest of HTML from this CGI script, will redirect to WEBDONE
-&print_cgihtml("");
+&send_working_trailer();
 
-close(WEBDONE_THTML);
-close(WEBWAIT_THTML);
 close(LOGFILE);
 close(MSGFILE);
 
@@ -630,48 +724,48 @@ sub show_status {
     return "&nbsp;&nbsp; $status_count . .<BR>\n";
 }
 #-----------------------------------
-
-sub check_marker {
-    # check if $str contains the given marker string (if $in marker != "")
-    my($str, $in) = @_;
-    $in && $str =~ /\b$in\b/;
-}
-#-----------------------------------
-
-sub replace_one {
-    # replaces $str in calling function (if $in marker != "")
-    my($str, $in, $out) = @_;
-    $in && $_[0] =~ s/\b$in\b/$out/;
-}
-#-----------------------------------
-
 sub replace_markers {
-    my($str) = @_;
-    &replace_one($str, $MRKR_HOSTNAME, "$hostname");
-    &replace_one($str, $MRKR_WEBDONEHTML, "$WORKHTTP/$WEBDONE_HTML");
-    &replace_one($str, $MRKR_INPUT, "$WORKHTTP/$SAVINPUT");
-    &replace_one($str, $MRKR_PDF, "$WORKHTTP/$WORKFILE.pdf");
-    &replace_one($str, $MRKR_GIF, "$WORKHTTP/$WORKFILE.gif");
-    &replace_one($str, $MRKR_POSTSCRIPT, "$WORKHTTP/$WORKFILE.ps");
-    &replace_one($str, $MRKR_MSGFILE, "$WORKHTTP/$MSGFILENAME");
-    &replace_one($str, $MRKR_TIMETAKEN, "$time_taken");
-    &replace_one($str, $MRKR_DONTREDIRECT, "$dontredirect");
-    return $str;
-}
-#-----------------------------------
+    my($tmpl) = @_;
+    # apply to only WORKING_TEMPLATE
+    $tmpl->param(MRKR_DONTREDIRECT => $dontredirect);
 
-sub print_cgihtml {
-    my($str) = @_;
-    # print lines in the CGI output file... until $str is encountered
-    # when completed this will finish the
-    # body, and do the javascript redirect to webdone.html
-    while (<WEBWAIT_THTML>) {
-	last if ($str && /$str/);
-	print &replace_markers($_);
+    # apply to both WORKING_TEMPLATE and DONE_TEMPLATE
+    $tmpl->param(MRKR_WEBDONEHTML => "$WORKHTTP/$WEBDONE_HTML");
+    $tmpl->param(MRKR_HOSTNAME => "$hostname ( $hostaddr )");
+    $tmpl->param(MRKR_ANYERROR => $errflag);
+    $tmpl->param(MRKR_MSGFILE => "$WORKHTTP/$MSGFILENAME");
+
+    # apply to only DONE_TEMPLATE
+    $tmpl->param(MRKR_INPUT => "$WORKHTTP/$SAVINPUT");
+    $tmpl->param(MRKR_PDF => "$WORKHTTP/$WORKFILE.pdf");
+    $tmpl->param(MRKR_MAKEGIF => $inmakegif);
+    $tmpl->param(MRKR_MAKEBCODE => $inbcodetext) if (-s "$BCODEFILE.eps");
+    $tmpl->param(MRKR_GIF => "$WORKHTTP/$WORKFILE.gif");
+    $tmpl->param(MRKR_POSTSCRIPT => "$WORKHTTP/$WORKFILE.ps");
+    $tmpl->param(MRKR_TIMETAKEN => $time_taken_units);
+    $tmpl->param(MRKR_FORM_HTML => $form_html);
+    $tmpl->param(MRKR_BCODE_EPS => "$WORKHTTP/$BCODEFILE.eps");
+
+    if ($errflag) {
+      if ($errcmd =~ /$CDLBL_E/) {
+        $tmpl->param(MRKR_CDLERROR => 1);
+      } elsif ($errcmd =~ /$PGM2GIF_E/) {
+        $tmpl->param(MRKR_GIFERROR=> 1);
+      } else {
+        $tmpl->param(MRKR_SYSERROR=> 1);
+      }
     }
-    # this completes the output HTML file (no need to print CGI::end_html)
 }
-#-----------------------------------
+
+#----------------------------------------
+sub send_working_trailer {
+    if (defined $working_trailer_tmpl) {
+	&replace_markers($working_trailer_tmpl);
+	$working_trailer_tmpl->output(print_to => *STDOUT);
+    }
+    # this completes the output HTML file displayed while processing job
+}
+#----------------------------------------
 
 sub do_cmd {
     my($cmd) = @_;
@@ -712,8 +806,6 @@ sub sighandler {
     print MSGFILE $str;
 
     close(WEBDONE_HTML);
-    close(WEBWAIT_THTML);
-    close(WEBDONE_THTML);
     close(LOGFILE);
     close(MSGFILE);
 
@@ -723,14 +815,15 @@ sub sighandler {
 #-----------------------------------
 
 sub myexit {
-    # if myexit() is called after WEBWAIT_THTML copied to STDOUT, 
+    # if myexit() is called after WORKING_TEMPLATE copied to STDOUT, 
     # the automatic redirect in that file needs to be suppressed
     $dontredirect = "1";
     $errflag++;
 
     my($mesg) = @_;
+    print "<html><head><title>Error</title></head><body>\n" unless ($sent_working_header);
     print "<P>\nSession ID# <B>$WORKDIRNAME</B> - Error - $mesg\n<HR><P>";
-    # displayed in middle of WEBWAIT_THTML
+    # displayed in middle of WORKING_TEMPLATE
 
     $datestr = localtime(time());
     $str = "$WORKDIRNAME [$datestr] Error - $mesg\n";
@@ -739,11 +832,9 @@ sub myexit {
     print MSGFILE $str;
 
     # complete rest of HTML from this CGI script
-    &print_cgihtml("");
+    if ($sent_working_header) { &send_working_trailer(); }
+    else { print "</body></html>\n"; }
 
-    close(WEBDONE_HTML);
-    close(WEBWAIT_THTML);
-    close(WEBDONE_THTML);
     close(LOGFILE);
     close(MSGFILE);
 
@@ -767,7 +858,7 @@ sub mktempdir {
 	last if ($done);
 	$n++;
     }
-    $done || myexit("mktempdir: mkdir in ($dir) with prefix ($try-$n) failed: $!");
+    $done || myexit("<p>mktempdir: mkdir in ($dir) with prefix ($try-$n) failed: $!\n<p>Try the job again, it may work next time.<p>");
     $done ? $dirname : "";
 }
 #-----------------------------------------------------------------------------
